"""
Module that summarizes an assignment group by conjunctive average.

Conjunctive average means: 

  1) We need all assignments to have a non-F grade.
  2) If there are A--F assignments present, we will compute the average of 
     those grades. For instance; an A and a C will result in a B; an A and a B 
     will result in an A, but an A with two Bs will become a B (standard 
     rounding).
"""

import datetime as dt

def summarize(user, assignments_list):
  """Extracts user's submissions for assignments in assingments_list to 
  summarize results into one grade and a grade date. Summarize by conjunctive 
  average."""

  pf_grades = []
  a2e_grades = []
  recent_date = dt.date(year=1970, month=1, day=1)

  for assignment in assignments_list:
    submission = assignment.get_submission(user)
    grade = submission.grade

    if grade is None:
      grade = "F"

    if grade in "ABCDE":
      a2e_grades.append(grade)
    else:
      pf_grades.append(grade)

    grade_date = submission.submitted_at or submission.graded_at

    if not grade_date:
      grade_date = recent_date
    else:
      grade_date = dt.date.fromisoformat(grade_date.split("T")[0])

    if grade_date > recent_date:
      recent_date = grade_date

  if not all(map(lambda x: x == "P", pf_grades)):
    return ("F", recent_date)

  if a2e_grades:
    return (a2e_average(a2e_grades), recent_date)
  return ("P", recent_date)
def a2e_average(grades):
  """Takes a list of A--E grades, returns the average."""
  num_grades = map(grade_to_int, grades)
  avg_grade = round(sum(num_grades)/len(grades))
  return int_to_grade(avg_grade)

def grade_to_int(grade):
  grade_map = {"E": 1, "D": 2, "C": 3, "B": 4, "A": 5}
  return grade_map[grade]

def int_to_grade(int_grade):
  grade_map_inv = {1: "E", 2: "D", 3: "C", 4: "B", 5: "A"}
  return grade_map_inv[int_grade]

def summarize_group(assignments_list, users_list):
  """Summarizes a particular set of assignments (assignments_list) for all
  users in users_list"""

  for user in users_list:
    grade, grade_date = summarize(user, assignments_list)
    yield (user, grade, grade_date)



import argparse
import canvasapi.course
import canvasapi.group
import canvaslms.cli
import canvaslms.cli.courses as courses
import canvaslms.hacks.canvasapi
import csv
import operator
import re
import sys

def add_users_command(subp):
  """Adds the users subcommand and its options to argparse subparser subp"""
  users_parser = subp.add_parser("users",
    help="Lists users of a course",
    description="Lists users of a course(s). Output, CSV-format: "
      "<course> [<group>] [<Canvas ID>] <login ID> [<LADOK ID>] "
      "<name> [<email>]")
  users_parser.set_defaults(func=users_command)
  courses.add_course_option(users_parser)
  add_group_option(users_parser)
  users_parser.add_argument("-s", "--students", action="store_true",
    help="Include students.")
  users_parser.add_argument("-a", "--assistants", action="store_true",
    help="Include teaching assistants.")
  users_parser.add_argument("-i", "--canvas-id", action="store_true",
    help="Include Canvas identifier")
  users_parser.add_argument("-l", "--ladok", action="store_true",
    help="Include LADOK identifier")
  users_parser.add_argument("-S", "--split-name", action="store_true",
    help="Returns first and last names as separate fields, "
      "instead of one containing the full name.")
  users_parser.add_argument("-e", "--email", action="store_true",
    help="Include email address")
  users_parser.add_argument("regex",
    default=".*", nargs="?",
    help="Regex for filtering users, default: '.*'")
def users_command(config, canvas, args):
  roles = []
  if args.students:
    roles.append("student")
  if args.assistants:
    roles.append("ta")
  output = csv.writer(sys.stdout, delimiter=args.delimiter)

  if args.group or args.category:
    rows = make_user_rows_w_groups(canvas, args, roles)
  else:
    rows = make_user_rows(canvas, args, roles)

  for row in rows:
    output.writerow(row)
def make_user_rows(canvas, args, roles):
  """Takes a list of courses and returns a list of users in those courses"""

  course_list = courses.process_course_option(canvas, args)

  users = filter_users(
    course_list,
    args.regex,
    roles)

  for user in users:
    try:
      row = []
      row.append(user.course.course_code)
      if args.canvas_id:
        row.append(user.id)
      try:
        row.append(user.login_id)
      except AttributeError:
        row.append(None)
      if args.ladok:
        try:
          row.append(user.integration_id)
        except AttributeError:
          row.append(None)
      if args.split_name:
        lastnames, firstnames = user.sortable_name.split(", ")
        row.append(firstnames.strip())
        row.append(lastnames.strip())
      else:
        row.append(user.name)
      if args.email:
        try:
          row.append(user.email)
        except AttributeError:
          row.append(None)
    except AttributeError as err:
      canvaslms.cli.warn(f"skipped {user}: {err}")
      continue

    yield row
def make_user_rows_w_groups(canvas, args, roles):
  """Takes a list of courses and returns a list of users in those courses,
  includes a group column"""

  groups = process_group_options(canvas, args)

  for group in groups:
    users = filter_users([group], args.regex, roles)

    for user in users:
      try:
        try:
          row = [group.category.course.course_code]
        except AttributeError:
          pass
        try:
          row = [group.course.course_code]
        except AttributeError:
          pass
        row.append(group.name)
        if args.canvas_id:
          row.append(user.id)
        try:
          row.append(user.login_id)
        except AttributeError:
          row.append(None)
        if args.ladok:
          try:
            row.append(user.integration_id)
          except AttributeError:
            row.append(None)
        if args.split_name:
          lastnames, firstnames = user.sortable_name.split(", ")
          row.append(firstnames.strip())
          row.append(lastnames.strip())
        else:
          row.append(user.name)
        if args.email:
          try:
            row.append(user.email)
          except AttributeError:
            row.append(None)
      except AttributeError as err:
        canvaslms.cli.warn(f"skipped {user}: {err}")
        continue

      yield row
def add_groups_command(subp):
  """Adds the groups subcommand and its options to argparse subparser subp"""
  groups_parser = subp.add_parser("groups",
    help="Lists groups of a course",
    description="Lists groups of a course(s). Output, CSV-format: "
      "<Canvas ID> <group category> <group name> <#members>")
  groups_parser.set_defaults(func=groups_command)
  courses.add_course_option(groups_parser)
  add_group_category_option(groups_parser)
  groups_parser.add_argument("regex", metavar="group_regex",
    default=".*", nargs="?",
    help="Regex for filtering groups, default: '.*'")
def groups_command(config, canvas, args):
  course_list = courses.process_course_option(canvas, args)
  output = csv.writer(sys.stdout, delimiter=args.delimiter)
  if args.category:
    categories = filter_group_categories(course_list, args.category)
  else:
    categories = filter_group_categories(course_list, ".*")

  for category in categories:
    for group in filter_groups([category], args.regex):
      row = [group.id, category.name, group.name, group.members_count]
      output.writerow(row)
def filter_group_categories(course_list, regex):
  """Filters out the group categories whose names match regex
  in the courses in course_list"""

  name = re.compile(regex or ".*")

  for course in course_list:
    for category in course.get_group_categories():
      if name.search(category.name):
        category.course = course
        yield category
def filter_groups(items, regex):
  """Items is a list of either courses or group categories,
  regex is a regular expression.
  Returns all groups whose name match regex."""

  name = re.compile(regex or ".*")

  for item in items:
    for group in item.get_groups():
      if name.search(group.name):
        if isinstance(item, canvasapi.course.Course):
          group.course = item
        elif isinstance(item, canvasapi.group.GroupCategory):
          group.category = item
        yield group
def list_users(courses, roles):
  """List users in courses with roles"""
  users = list()

  for course in courses:
    course_users = list(course.get_users(enrollment_type=roles))
    for user in course_users:
      user.course = course
    users.extend(course_users)

  return users
def filter_users(course_list, regex, roles=[]):
  """
  Filter users in courses with roles based on regex. `regex` is matched on 
  - Canvas ID (exact match),
  - name (regex),
  - login ID (regex),
  - integration id (exact match),
  - SIS user ID (exact match).
  """
  pattern = re.compile(regex or ".*")

  for user in list_users(course_list, roles):
    if str(user.id) == regex:
      yield user
      continue

    if pattern.search(user.name):
      yield user
      continue

    try:
      if pattern.search(user.login_id):
        yield user
        continue
    except AttributeError:
      canvaslms.cli.warn(f"{user} has no login_id")

    try:
      if user.integration_id == regex:
        yield user
        continue
    except AttributeError:
      canvaslms.cli.warn(f"{user} has no integration_id")

    try:
      if user.sis_user_id == regex:
        yield user
        continue
    except AttributeError:
      canvaslms.cli.warn(f"{user} has no sis_user_id")
def list_students(courses):
  """List users in courses with role student"""
  return list_users(courses, ["student"])

def list_teachers(courses):
  """List users in courses with role teacher"""
  return list_users(courses, ["teacher"])
def get_uid(user):
  """
  Takes a user object and returns a unique identifier.

  Returns one of login_id, integration_id, sis_user_id, id (Canvas ID); in that 
  order of preference.
  """
  attributes = ["login_id", "integration_id", "sis_user_id", "id"]
  for attribute in attributes:
    try:
      return operator.attrgetter(attribute)(user)
    except AttributeError:
      pass

  raise AttributeError(f"no unique user attribute existed, tried: {attributes}")
def add_user_option_wo_depends(parser, required=False):
  """
  Adds the -u option to argparse parser,
  without adding other required options
  """
  help="Filter users on name, login ID or Canvas ID by user_regex"
  options = {"required": required}
  if not required:
    options["default"] = ".*"
    help += ", default: '.*'"
  else:
    help += ", required: use '.*' to match all students"

  parser.add_argument("-u", "--user", metavar="user_regex",
    help=help, **options)

def add_user_option(parser, required=False):
  """Adds the -u option to argparse parser"""
  try:
    courses.add_course_option(parser)
  except argparse.ArgumentError:
    pass

  add_user_option_wo_depends(parser, required)
def process_user_option(canvas, args):
  """Processes the user option from command line, returns a list of users"""
  return list(filter_users(
    courses.process_course_option(canvas, args),
    args.user))
def add_group_category_option_wo_depends(parser):
  """Adds the group category option, without adding required other options"""
  parser.add_argument("-C", "--category", metavar="category_regex",
    required=False,
    help="Filters groups only from the group categories matching "
      "category_regex")

def add_group_category_option(parser):
  """Adds the group category option, adds required options"""
  try:
    courses.add_course_option(parser)
  except argparse.ArgumentError:
    pass

  add_group_category_option_wo_depends(parser)

def add_group_option_wo_depends(parser):
  """
  Adds group filtering option to argparse parser,
  without adding required options
  """
  try:
    add_group_category_option_wo_depends(parser)
  except argparse.ArgumentError:
    pass

  parser.add_argument("-G", "--group", metavar="group_regex",
    required=False,
    help="Filters groups whose name match group_regex")

def add_group_option(parser):
  """Adds group filtering options to argparse parser,
  adds required other options"""
  try:
    courses.add_course_option(parser)
  except argparse.ArgumentError:
    pass

  add_group_option_wo_depends(parser)
def process_group_options(canvas, args):
  """Processes the group/group category options, returns a list of groups"""

  course_list = courses.process_course_option(canvas, args)

  if args.category:
    return list(filter_groups(
      filter_group_categories(course_list, args.category),
      args.group))

  return list(filter_groups(
    course_list,
    args.group))
def add_user_or_group_option(parser, required=False):
  """Adds user and group options as mutually exclusive options to parser"""
  try:
    courses.add_course_option(parser)
  except argparse.ArgumentError:
    pass

  parser = parser.add_mutually_exclusive_group(required=required)

  try:
    add_user_option_wo_depends(parser)
  except argparse.ArgumentError:
    pass

  try:
    add_group_option_wo_depends(parser)
  except argparse.ArgumentError:
    pass

def process_user_or_group_option(canvas, args):
  """Returns a list of users, filtered either by user regex or by groups"""
  if args.group or args.category:
    users = list()
    for group in process_group_options(canvas, args):
      users.extend(group.get_users())

    return users

  return process_user_option(canvas, args)

def add_command(subp):
  """Adds the subcommands users and groups to argparse subparser subp"""
  add_users_command(subp)
  add_groups_command(subp)



import canvasapi
import json
import os

canvas = canvasapi.Canvas(os.environ["CANVAS_SERVER"],
        os.environ["CANVAS_TOKEN"])

test_course = None

for course in canvas.get_courses():
    if course.name == "Sandbox dbosk":
        test_course = course
        break

test_quiz = None

for quiz in course.get_quizzes():
    if quiz.title == "Classic datorprov":
        test_quiz = quiz
        break

test_submission = None

# There is only one in my setup
for quiz_submission in test_quiz.get_submissions():
    test_submission = quiz_submission

print("\n# Quiz submission questions\n")

for subm_question in test_submission.get_submission_questions():
    #print(subm_question.__dict__.keys())
    #print(subm_question)
    print(f"{subm_question.id} "
          f"{subm_question.question_name}:\n"
          f"{subm_question.question_text}")
    try:
        print(f"Alternatives: {subm_question.answers}")
        print(f"Correct: {subm_question.correct}")
    except AttributeError:
        pass

    print()

print("\n# Quiz submission answers\n")

quiz = None

for assignment in test_course.get_assignments():
    if assignment.name == "Classic datorprov":
        quiz = assignment
        break

for submission in quiz.get_submissions(include=["submission_history"]):
    for subm in submission.submission_history:
        #print(subm)
        try:
            for data in subm["submission_data"]:
                print(json.dumps(data, indent=2))
        except KeyError:
            pass




"""A command-line interface for the Canvas LMS."""

import appdirs
import argcomplete, argparse
from canvasapi import Canvas
import canvaslms.cli.login
import json
import os
import pathlib
import sys

import canvaslms.cli.login
import canvaslms.cli.courses
import canvaslms.cli.users
import canvaslms.cli.assignments
import canvaslms.cli.submissions
import canvaslms.cli.grade
import canvaslms.cli.results

dirs = appdirs.AppDirs("canvaslms", "dbosk@kth.se")

def err(rc, msg):
  """Prints msg to stderr, prints a stack trace and
  exits with rc as return code"""
  print(f"{sys.argv[0]}: error: {msg}", file=sys.stderr)
  sys.exit(rc)

def warn(msg):
  """Prints msg to stderr"""
  print(f"{sys.argv[0]}: {msg}", file=sys.stderr)
def read_configuration(config_file):
  """Returns a dictionary containing the configuration"""
  config = {}

  try:
    with open(config_file, "r") as file:
      config.update(json.load(file))
  except FileNotFoundError:
    pass
  except json.decoder.JSONDecodeError as err:
    warn(f"config file is malformed: {err}")
  if "canvas" not in config:
    config["canvas"] = {}

  if "CANVAS_SERVER" in os.environ:
    config["canvas"]["host"] = os.environ["CANVAS_SERVER"]

  if "CANVAS_TOKEN" in os.environ:
    config["canvas"]["access_token"] = os.environ["CANVAS_TOKEN"]

  return config
def update_config_file(config, config_file):
  """Updates the config file by writing the config dictionary back to it"""
  try:
    with open(config_file, "w") as fd:
      json.dump(config, fd)
  except FileNotFoundError:
    os.makedirs(pathlib.PurePath(config_file).parent)
    with open(config_file, "w") as fd:
      json.dump(config, fd)

def main():
  argp = argparse.ArgumentParser(
    description="Scriptable Canvas LMS",
    epilog="Web: https://github.com/dbosk/canvaslms")

  subp = argp.add_subparsers(
    title="commands",
    dest="command",
    required=True)

  argp.add_argument("-f", "--config-file",
    default=f"{dirs.user_config_dir}/config.json",
    help="Path to configuration file "
      f"(default: {dirs.user_config_dir}/config.json) "
      "or set CANVAS_SERVER and CANVAS_TOKEN environment variables.")
  argp.add_argument("-d", "--delimiter",
    default="\t",
    help="Sets the delimiter for CSV output, the default is the tab character")
  canvaslms.cli.login.add_command(subp)
  canvaslms.cli.courses.add_command(subp)
  canvaslms.cli.users.add_command(subp)
  canvaslms.cli.assignments.add_command(subp)
  canvaslms.cli.submissions.add_command(subp)
  canvaslms.cli.grade.add_command(subp)
  canvaslms.cli.results.add_command(subp)

  argcomplete.autocomplete(argp)
  args = argp.parse_args()

  config = read_configuration(args.config_file)

  if args.func != canvaslms.cli.login.login_command:
    hostname, token = canvaslms.cli.login.load_credentials(config)

    if not (hostname and token):
      err(1, "No hostname or token, run `canvaslms login`")

    if "://" not in hostname:
      hostname = f"https://{hostname}"

    canvas = Canvas(hostname, token)
  else:
    canvas = None

  if args.func:
    args.func(config, canvas, args)
